#include <Arduino.h>

#ifdef ESP32
  #include <WiFi.h>
  int BAUD_RATE = 115200;
  
#elif defined(ARDUINO_UNOR4_WIFI)
    #include <WiFiS3.h>
    int BAUD_RATE = 9600;
#else
  #include <ESP8266WiFi.h>
  int BAUD_RATE = 9600;
#endif

/////// The code below contains various examples, simply remove the things you don't need //////

////// ***** ChipChop definitions  ***** ///////
#include <ChipChopManager.h> 
ChipChopManager ChipChop;

//The connection credentials can be found in the device settings Dev Console > Devices
String server_uri = "...your allocated Endpoint Server URI..."; 
String uuid = "...your ChipChop UUID ..."; 
String auth_code = "... your Security Authentication Code ..."; 
String device_id = "... your Device ID ..."; // same id as you have specified in the Dev Console for this device, make sure that this is unique between all your devices

// ChipChop callback function triggered every time you send a command to your device either using Alexa, Web Viewer App, Development Console or through an automated Action
void ChipChop_onCommandReceived(String target_component,String command_value, String command_source, int command_age){
    Serial.print("command age => ");
    Serial.print(command_age);
    Serial.println("");
    Serial.println("target component => " + target_component);
    Serial.println("command source => " + command_source);
    Serial.println("command value => " + command_value);

    // "command_age" is the difference (in seconds) between the current time and the time when this command was received on ChipChop. If the command was triggered when the device was not connected it will be forwarded on the next succesfully device heartbeat. It's up to you to decide how "old" a command can be and should you act on it. In optimal conditions the command age will be close or equal to 0 (zero)
    // This is just a convenience feature and can be safely ignored by removing the "if" statement below
    if(command_age > 60){ // command is more than 60 sec old 
        return; // ignore the command, it's too old
    }

    // "target_component" is the name of a components that should act on this command. If the command is intended for the entire device or was sent to a device group the target_component will be "any".

    // IMPORTANT: "command_value" is always passed to this function as a String, if a certain component is expecting an integer or a float you need to cast the command_value to the correct type, see examples below.

    //EXAMPLES:

    //currently the "command source" (if it's not "undefined") is telling us which device has triggered this event through an "Action"
      //in the future releases it may also show if the command was issued through the Dev Console, Web App or Alexa
    if(command_source == "some_device_id"){
        //do something ....
    }

    // EXAMPLE: sent to the entire device
    if(target_component == "any"){
      //do something here with the "command_value"....
      
    }

    //EXAMPLE: command sent to a component called "led_1" with a String value "ON" or "OFF"
    if(target_component == "led_1"){
        if(command_value == "ON"){
            //do something to turn the led on 
            // digitalWrite(led_pin,LOW);

        }else if(command_value == "OFF"){
            //do something to turn the led off 
            // digitalWrite(led_pin,HIGH);
        }

    }

    //EXAMPLE:  command sent to a component called "brightness" that expects an integer value
    if(target_component == "brightness"){ 
        // do something here to change the brightness to the received percentage....
        int brightness_percent = command_value.toInt(); // or .toFloat() or .toDouble()

    }

    //IMPORTANT: You should always immediately update any changed component status unless the target_component was "any"
    ChipChop.updateStatus(target_component,command_value); // Confirm the LED status change

}

// ChipChop callback function triggered if an error message is received
void ChipChop_onError(String error){
    Serial.println("ChipChop error => " + error);
}

////// ***** END of ChipChop definitions  ***** ///////

unsigned long loopCycle = 0; //used in the main loop to avoid using blocking delay()

void setup()
{

  Serial.begin(BAUD_RATE);

  WiFi.begin("< your SSID >", "< your wifi password >");

    Serial.print("Connecting");
    while (WiFi.status() != WL_CONNECTED)
    {
      delay(500);
      Serial.print(".");
    }


  //*** ChipChop utility callbacks, functions & settings ***//
  ChipChop.commandCallback(ChipChop_onCommandReceived); // remote command callback function defined before setup()
  ChipChop.errorCallback(ChipChop_onError); // error message callback function defined before setup()

  ChipChop.start(server_uri, uuid, device_id, auth_code); // IMPORTANT: initialise and start here the ChipChop engine 

  //*** OPTIONAL ***///
  ChipChop.heartBeatInterval(10); //set the heartbeat interval in seconds.Smallest allowed value is 10 seconds (default). Reducing this interval to less that 10 sec will result in initial warnings then connection being closed for a time period and eventually a full account suspension. If your device's status doesn't change often you can safely increase this value to a lot longer period.

  ChipChop.requestRate(550); //in ms...You can set the maximum speed that trigger events can be sent. It helps to prevent uncontrollable bursts of events. The min value is 500ms.

  ChipChop.keepAlive(1,120000); // (0) keep alive is off. When set to (1) monitors the socket connection and if a loss of connection is detected for a predefined period it will attempt to clear & restart the websockets from memory. Setting keep alive to (2) is ESP only and will fully restart the ESP. The second parameter is the predefined timeout value in milliseconds, default is 120000ms (2 minutes)
  ChipChop.restart(); //requires keepAlive to be set. It will call immediately the restart routine based on the keepAlive setting without waiting for a timeout.

  ChipChop.debug(true); // true/false enables log messages from the ChipChop library
  ChipChop.useClock(true); // use the real time ChipChop server clock ...see example use in the main loop below
  ChipChop.setDateFormat("dd MMM YY - hh:mm:ss"); // YY - year 2 digits, YYYY - year 4 digits, MM - month number, MMM - month short string, MMMM - month long string, dd - day as number , hh - hour, mm - minute, ss - second

  //*** end of ChipChop utility callbacks and functions ***//

}


void loop(){

  ChipChop.run(); // IMPORTANT: This call has to be here in the main loop() to keep ChipChop engine running continuously otherwise you will lose your real-time connection channel (web socket)


  // you should avoid using delay() in your code as it can block the real-time socket communication
  // instead use non-blocking approach, for example here we are using the loopCycle variable (declared just before the setup() ) to 
  // execute the code very 10 seconds (10000 milliseconds)
  if ((millis() - loopCycle) > 10000){

      if (WiFi.status() == WL_CONNECTED){ //no point sending anything if we are not connected to the WiFi

        // check if a connection with ChipChop is fully established
        // this property will change to true (1) only after a socket connection is established and the handshake procedure is successfully completed
        // it will turn to false on connection loss or ChipChop explicitly closing the connection for system violation reasons
        bool connection_established = ChipChop.connected; 

        // Update status example
        ChipChop.updateStatus("led","ON"); // status as string
        ChipChop.updateStatus("temperature",26); // status as integer
        ChipChop.updateStatus("humidity",float(-44.567)); // status as float

        // Send a trigger event example
        ChipChop.triggerEvent("motion_sensor","DETECTED"); //as string
        ChipChop.triggerEvent("water_level",42); // as integer
        ChipChop.triggerEvent("proximity",float(33.05)); // as float

        // Remove a component from the component list (the list is automatically generated each time you use ChipChop.updateStatus("some_component_name"))
        // This will effectively stop sending a status update for a specific component until you use again ChipChop.updateStatus("some_component_name")
        // This is useful if you want to capture the last time some component state has changed
        ChipChop.removeStatus("motion_sensor"); 

        //*** ChipChop Real Time Clock various calls ***///
        // for this to work correctly you should have used in the setup "ChipChop.useClock(true)" and your device has to be succesfully connected to ChipChop
        Serial.println(ChipChop.Date.year);
        Serial.println(ChipChop.Date.month); // 1, 2, 3, 4
        Serial.println(ChipChop.Date.monthLong); // Jan, Feb, Mar ....
        Serial.println(ChipChop.Date.monthShort); // January, February, March...
        Serial.println(ChipChop.Date.day); // Day 1 = Monday and day 7 = Sunday

        Serial.println(ChipChop.Date.hour); //in 24 hrs format
        Serial.println(ChipChop.Date.minute);
        Serial.println(ChipChop.Date.second);

        Serial.println(ChipChop.Date.timestamp); // the current UNIX timestamp in seconds

        // get automatically formatted date/time string, useful to display on lcd/oled screens
        // set the format first in the setup() above using ChipChop.setDateFormat() returns date & time String like: 01 Jan 2023 - 18:45:32
        Serial.println(ChipChop.Date.formatted); 
        
      }

    loopCycle = millis();
  }

}

