#include <Arduino.h>

#ifdef ESP32
  #include <WiFi.h>
  int BAUD_RATE = 115200;
#elif defined(ARDUINO_UNOR4_WIFI)
    #include <WiFiS3.h>
    int BAUD_RATE = 9600;
#else
  #include <ESP8266WiFi.h>
  int BAUD_RATE = 9600;
#endif

////// ***** ChipChop definitions  ***** ///////
#include <ChipChopManager.h> 
ChipChopManager ChipChop;

//The connection credentials can be found in the device settings Dev Console > Devices
String server_uri = "...your allocated Endpoint Server URI..."; 
String uuid = "...your ChipChop UUID ..."; 
String auth_code = "... your Security Authentication Code ..."; 
String device_id = "... your Device ID ..."; // same Unique ID as you have specified in the Dev Console for this device, make sure that this is unique between all your devices

////// ***** Led variables ***** //////
String button_status = "OFF"; // the device will start with the button in the OFF state
const int button_pin = 14; // define the pin that will listen for the button being pressed, we are using here the ESP pin D5 which is an Arduino pin 14 but can be any digital pin
bool can_change_state = true; // used to implement a toggle button behavior
unsigned long last_state_change_time = 0; // used to prevent sending trigger events too quickly

// ChipChop callback function triggered if an error message is received
void ChipChop_onError(String error){
    Serial.println("ChipChop error => " + error);
}

void setup()
{

    Serial.begin(BAUD_RATE);

    WiFi.begin("< your SSID >", "< your wifi password >");

    Serial.print("WiFi Connecting");
    while (WiFi.status() != WL_CONNECTED)
    {
        delay(500);
        Serial.print(".");
    }
  
    //*** declare ChipChop utility callbacks, functions & settings ***//
    ChipChop.start(server_uri, uuid, device_id, auth_code);
    ChipChop.debug(true); // enables log messages from the ChipChop library, set to false in production
    ChipChop.errorCallback(ChipChop_onError); // error message callback function defined before setup()

    // declare the button pin as input, note that the LOW / HIGH states are in this case logically inverted i.e. LOW == button is pressed (ON), HIGH == button is not pressed (OFF)     
    pinMode(button_pin,INPUT_PULLUP);

    ChipChop.updateStatus("button",button_status); // You should always set the initial status of all components 

}


void loop(){


    ChipChop.run(); // This call has to be here in the main loop() and needs to run continously

    /*
    We will be constantly listening for the button press. The same would work for a motion sensor being triggered, a touch button or pretty much anything that changes a digital pin to be HIGH or LOW

    IMPORTANT: We have to be careful if we are using anything that instantly changes it's state from LOW to HIGH or HIGH to LOW (like a momentary switch or motion sensor) as:
    1. the button press/release would fire 2 trigger events in quick succession and we have a hard limit of maximum 10 events a minute so we can run out of events
    2. the changes in the button state can happen too quickly, ChipChop will accept a max of 1 event per 500-600 milliseconds anything faster will be rejected 
    
    To avoid system violations and potential temporary block of the communication channel by the ChipChop server we should consider the following:
    a. Report the OFF state through the heartbeat as it would anyway happen at some point within the next 10 seconds or leave at least couple of seconds gap between the HIGH/LOW (ON/OFF) events
    b. Avoid at all costs firing a continuous flood of events, for example if a motion sensor is detecting motion we should only send the first event (when motion is first detected) and only send another one when the motion has completely stopped. Most motion sensors will usually have a hardwired block between detections but push buttons don't.
    c. If we've set another (remote) device to an ON state when a button on this device is pressed, for example a "Doorbell Button" device sets a "Warning Light" device to ON, reseting the Warning Device (turning the warning light OFF) should be handled by the Warning Device itself and not by the Doorbell Button being released
    d. Where possible, with buttons implement a "toggle" behavior where we ignore the release i.e. first button press sets the state to ON, next press toggles to OFF , next toggles to ON etc...

    */
    
    // We are implementing a button "toggle" behavior so only the button pressed (LOW) state is relevant and we are ignoring the continuous press of the button. We will also track the time between clicks (last_state_change_time) to avoid firing trigger events too quickly
    if(digitalRead(button_pin) == LOW){ //button is pressed
        
        if(can_change_state == true){ //ignore the continuous press, only when the button is released this changes to true
            if(button_status == "OFF"){ 
                button_status = "ON";
            }else if(button_status == "ON"){ 
                button_status = "OFF";
            }

            can_change_state = false; //on the next loop ignore if the button is still pressed
            last_state_change_time = millis(); //cache the current change of state time

            ChipChop.triggerEvent("button",button_status); // send a Trigger Event to ChipChop

        }
    }else if(digitalRead(button_pin) == HIGH){ //button is not pressed
        
        if(can_change_state == false){
            if(millis() - last_state_change_time > 500){ // only execute if the time passed from the last button press is over 500ms
                can_change_state = true; //permit the change on the next button press
            }
        }
        
    }

    // Update the button status continuously, the ChipChop library will ensure that they are sent in correct 10 seconds intervals
    ChipChop.updateStatus("button",button_status);
 

}

