/****************************************************************************************************************************
  ChipChopManager.h
  For ChipChop Arduino Library

  This library handles all the communication with ChipChop.io IoT engine automatically and exposes some useful little utilities that you can call from your main sketch or any other class.

  Built by ChipChop Gizmo - https://chipchop.io/api_docs/chipchop-arduino-library 
  Licensed under GNU license

  Version 1.27

  Version Modified By   Date      Comments
  ------- -----------  ---------- -----------
  1.00    Gizmo        24/05/2022 Initial build to incorporate a number of smaller classes & utilities into a single portable class 
  1.24    Gizmo        25/06/2023 Changed to using Websockets2 Generic library (from ArduinoWebsockets 0.5.3) and provides a cut-down version for ESP8266/ESP32
  1.34    Gizmo        25/08/2023 Includes the WebSocket support for UNO R4 WiFi in the provided "Websockets2 Generic - Modified" library
                                  Improved initial connection handshake procedure to avoid long waits and system penalties when attempting reconnections during testing

  1.36    Gizmo        04/11/2023 Added the option to explicitly close the connection ChipChop
                                  Added the option to remove a component from the heartbeat so it's status won't be sent
                                  Added ChipChop.connected property

  1.40    Gizmo        17/02/2024 Library event management reworked 
                                  keepAlive() added - option 2 is ESP Only
                                  requestRate() added
                                  restart() added


 *****************************************************************************************************************************/


#ifndef CHIPCHOP_H
#define CHIPCHOP_H

 #include <Arduino.h>


class ChipChopManager{

    using commandCallbackType = void (*)(String,String,String,int);
    using errorCallbackType = void (*)(String);
 
    private:

        String  _SERVER = "";

        const String _quote = "\"";
        
        typedef struct{
            String key;
            String value;
        } _JSON;

        typedef struct{
            String event;
            String status;
        } _EVENT;

        _EVENT _send_buffer[11];
        int buffer_size = 0;
        unsigned long _send_rate = 510;
        unsigned long _send_pulse_timer = 0;
        void sendNextEvent();
        void addEventToBuffer(String type, String status);
        void popBufferQueue(int poz);
        void limitPopBufferQueue(String event, String status);
        void clearBufferQueue();

        bool _LOG_ENABLED = false;
        bool _LOG_RESPONSE_ONLY = false;
        
        String _UUID;
        String _DEVICE_ID;
        String _AUTH_CODE;

        unsigned long lastPollTime = 0;

        int _keep_alive = 0; // 0. off, 1. monitor wifi only and pause 2. aggressive esp restart
        String _restart_reason = "Restart happened after initial wakeup";
        unsigned long _last_chipchop_connection_ok = 0;
        unsigned long _keep_alive_timeout = 120000; //2 mins default, minimum 30sec max 10 minutes
        int _curr_connection_status = 0;
        void keepAliveCheck();
        void restartSockets();


       
        unsigned long _HEARTBEAT_INTERVAL = 10000; // setting to 0 changes the device to a "status request" only mode
        unsigned long lastHeartbeatSendTime = 0;
        unsigned long lastDateTimeRefresh = 0;
        bool _can_send_heartbeat = 1;

        unsigned long eventSpacing = 0;
        unsigned long lastEventTime = 0;
        byte eventsSent = 0;

        unsigned long _SERVER_TIME = 0; // received server time in millis used to run a real time clock
        unsigned long _DEVICE_TIME_MILLIS = 0;
        
        int _MAX_COMPONENTS = 10;
        _JSON _COMPONENTS[10];

       
       void _respond_to_status_request(const String& message_id);
       void _302_api_redirect(const String& api_uri);

        bool socket_restart_mode = 0;
        unsigned long socket_restart_timer = 0;
        int socket_restart_timeout = 0;

        void _send_heartbeat();
        void _validate_trigger_event(String component_name);
        void _send_trigger_event(String status);

        void _connect_socket();

        String zeroPrefix(int num);

        void setComponentStatus(const String component, String value);
        String componentsToJson(String triggerEventComponent);

        unsigned long stringToLong(String input);
        float stringToFloat(String input);

        commandCallbackType commandReceivedCallback;
        bool hasCommandCallback = false;
        

        void sendCommandCallback(String target, String value, String source, int command_age){
            if(hasCommandCallback == true){
                commandReceivedCallback(target, value, source, command_age);
            }
        }

        errorCallbackType errorReceivedCallback;
        bool hasErrorCallback = false;
                
        void sendErrorCallback(String error){
            if(hasErrorCallback == true){
                errorReceivedCallback(error);
            }
        }

        

        void setTime();
        bool _use_clock = false;

        ///The basic MD5 functions
        typedef unsigned long MD5_u32plus;

        typedef struct {
            MD5_u32plus lo, hi;
            MD5_u32plus a, b, c, d;
            unsigned char buffer[64];
            MD5_u32plus block[16];
        } MD5_CTX;

        static unsigned char* MD5_make_hash(char *arg);
        static unsigned char* MD5_make_hash(char *arg,size_t size);
        static char* MD5_make_digest(const unsigned char *digest, int len);
        static const void *MD5_body(void *ctxBuf, const void *data, size_t size);
        static void MD5Init(void *ctxBuf);
        static void MD5Final(unsigned char *result, void *ctxBuf);
        static void MD5Update(void *ctxBuf, const void *data, size_t size);

    public:

        void _process_server_message(const String& data); //has to be set as "public" due to the WebSocket class implementation but should not be called from anywhere and should be considered private
        unsigned long lastReconnectTime = 0;
        bool connected = false;
        bool handshakeComplete = 0;

        ChipChopManager();
        void run();

        void start(String server_uri, String uuid, String device_id, String auth_code );
        void closeConnection();
        void debug(bool enabled){
            _LOG_ENABLED = enabled;
        }
        void debug_resp_only(bool enabled){
            _LOG_ENABLED = enabled;
            _LOG_RESPONSE_ONLY = enabled;
        }

        void requestRate(int rate){
            if(rate < 500){
                rate = 500;
            }
            _send_rate = rate;
        }

        void log(String data);
        void restart();
        void keepAlive(int mode, int timeout = 120000){
            _keep_alive = mode;
            if(timeout < 30000){
                _keep_alive_timeout = 30000;
            }else if (timeout > 600000){
                _keep_alive_timeout = 600000;
            }else{
                _keep_alive_timeout = timeout;
            }
        }

        void commandCallback(commandCallbackType cb){
            commandReceivedCallback = cb;
            hasCommandCallback = true;
        }
        void errorCallback(errorCallbackType cb){
            errorReceivedCallback = cb;
            hasErrorCallback = true;
        }

        String JSON_find(const String& obj, const String& property);
        String JSON_stringify(_JSON obj[], int obj_length);

        // set the heartbeat interval in seconds, the minimum interval is 10 seconds and anything faster
        // will incur system violations marked on the account and potential dropping of the communication channel
        void heartBeatInterval(int interval){
            if(interval < 10){
                interval = 10;
            }
            _HEARTBEAT_INTERVAL = interval * 1000; //convert to milliseconds
        }

        void triggerEvent(String component_name,String status);
        void triggerEvent(String component_name,int status);
        void triggerEvent(String component_name,float status);

        void updateStatus(String component_name,String status);
        void updateStatus(String component_name,int status);
        void updateStatus(String component_name,float status);

        void removeStatus(String component_name);

        struct time_struct{
            int year;
            int yearShort;
            int month;
            int day;
            int hour;
            int minute;
            int second;
            unsigned long timestamp; //UNIX timestamp in seconds
            String monthLong;
            String monthShort;
            String formatted;
        };

        time_struct Date;
        String dateFormat = "dd MM YY - hh:mm:ss";
        void setDateFormat(String format){
            dateFormat = format;
        };

        void useClock(bool use){
            _use_clock = use;
        }

};

extern ChipChopManager ChipChop;

#endif



